@@ -9,8 +9,6 @@
 
 #include "avif/avif.h"
 
-#include "testcase.h"
-
 #include <inttypes.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -95,165 +93,6 @@ static const char * nextFilename(const char * parentDir, const char * extension,
 }
 #endif
 
-static int generateEncodeDecodeTests(const char * dataDir)
-{
-    printf("AVIF Test Suite: Generating Encode/Decode Tests...\n");
-
-    int retCode = 0;
-    cJSON * tests = cJSON_CreateArray();
-
-    struct QuantizerPairs
-    {
-        int minQP;
-        int maxQP;
-    } quantizerPairs[] = {
-        { 0, 0 },  // lossless
-        { 4, 40 }, // Q60
-        { 24, 60 } // Q40
-    };
-    const int quantizerPairsCount = sizeof(quantizerPairs) / sizeof(quantizerPairs[0]);
-
-    NextFilenameData nfd;
-    memset(&nfd, 0, sizeof(nfd));
-    const char * filename = nextFilename(dataDir, "y4m", &nfd);
-    for (; filename != NULL; filename = nextFilename(dataDir, "y4m", &nfd)) {
-        avifCodecChoice encodeChoices[] = { AVIF_CODEC_CHOICE_AOM, AVIF_CODEC_CHOICE_RAV1E };
-        const int encodeChoiceCount = sizeof(encodeChoices) / sizeof(encodeChoices[0]);
-        for (int encodeChoiceIndex = 0; encodeChoiceIndex < encodeChoiceCount; ++encodeChoiceIndex) {
-            avifCodecChoice decodeChoices[] = { AVIF_CODEC_CHOICE_AOM, AVIF_CODEC_CHOICE_DAV1D, AVIF_CODEC_CHOICE_LIBGAV1 };
-            const int decodeChoiceCount = sizeof(decodeChoices) / sizeof(decodeChoices[0]);
-            for (int decodeChoiceIndex = 0; decodeChoiceIndex < decodeChoiceCount; ++decodeChoiceIndex) {
-                for (int qpIndex = 0; qpIndex < quantizerPairsCount; ++qpIndex) {
-                    int speeds[] = { AVIF_SPEED_DEFAULT, 10 };
-                    int speedCount = sizeof(speeds) / sizeof(speeds[0]);
-                    for (int speedIndex = 0; speedIndex < speedCount; ++speedIndex) {
-                        TestCase * tc = testCaseCreate();
-                        testCaseSetInputFilename(tc, filename);
-                        tc->encodeChoice = encodeChoices[encodeChoiceIndex];
-                        tc->decodeChoice = decodeChoices[decodeChoiceIndex];
-                        tc->active = AVIF_TRUE;
-                        tc->speed = speeds[speedIndex];
-                        tc->minQuantizer = quantizerPairs[qpIndex].minQP;
-                        tc->maxQuantizer = quantizerPairs[qpIndex].maxQP;
-                        testCaseGenerateName(tc);
-
-                        if (!testCaseRun(tc, dataDir, AVIF_TRUE)) {
-                            printf("ERROR: Failed to run test case: %s\n", tc->name);
-                            goto cleanup;
-                        }
-
-                        cJSON_AddItemToArray(tests, testCaseToJSON(tc));
-                        testCaseDestroy(tc);
-                    }
-                }
-            }
-        }
-    }
-
-    char * jsonString = cJSON_PrintUnformatted(tests);
-
-    char testJSONFilename[2048];
-    snprintf(testJSONFilename, sizeof(testJSONFilename), "%s/tests.json", dataDir);
-    testJSONFilename[sizeof(testJSONFilename) - 1] = 0;
-    FILE * f = fopen(testJSONFilename, "wb");
-    if (f) {
-        fprintf(f, "%s", jsonString);
-        fclose(f);
-
-        printf("Wrote: %s\n", testJSONFilename);
-    } else {
-        printf("Failed to write: %s\n", testJSONFilename);
-        retCode = 1;
-    }
-    free(jsonString);
-
-cleanup:
-    cJSON_Delete(tests);
-    return retCode;
-}
-
-static int runEncodeDecodeTests(const char * dataDir, const char * testFilter)
-{
-    printf("AVIF Test Suite: Running Encode/Decode Tests...\n");
-
-    char testJSONFilename[2048];
-    snprintf(testJSONFilename, sizeof(testJSONFilename), "%s/tests.json", dataDir);
-    testJSONFilename[sizeof(testJSONFilename) - 1] = 0;
-    FILE * f = fopen(testJSONFilename, "rb");
-    if (!f) {
-        printf("ERROR: Failed to read: %s\n", testJSONFilename);
-        return 1;
-    }
-
-    fseek(f, 0, SEEK_END);
-    size_t fileSize = ftell(f);
-    fseek(f, 0, SEEK_SET);
-
-    char * rawJSON = malloc(fileSize + 1);
-    if (fread(rawJSON, 1, fileSize, f) != fileSize) {
-        printf("ERROR: Failed to read: %s\n", testJSONFilename);
-        free(rawJSON);
-        fclose(f);
-        return 1;
-    }
-    rawJSON[fileSize] = 0;
-    fclose(f);
-
-    cJSON * tests = cJSON_Parse(rawJSON);
-    if (!tests || !cJSON_IsArray(tests)) {
-        if (tests) {
-            cJSON_Delete(tests);
-        }
-        printf("ERROR: Invalid JSON: %s\n", testJSONFilename);
-        return 1;
-    }
-    free(rawJSON);
-
-    int totalCount = 0;
-    int skippedCount = 0;
-    int failedCount = 0;
-    for (cJSON * t = tests->child; t != NULL; t = t->next) {
-        if (!cJSON_IsObject(t)) {
-            ++skippedCount;
-            continue;
-        }
-
-        TestCase * tc = testCaseFromJSON(t);
-        if (!tc || !tc->active) {
-            ++skippedCount;
-            testCaseDestroy(tc);
-            continue;
-        }
-
-        if (testFilter) {
-            if (strstr(tc->name, testFilter) == NULL) {
-                ++skippedCount;
-                testCaseDestroy(tc);
-                continue;
-            }
-        }
-
-        // Skip the test if the requested encoder or decoder is not available.
-        if (!avifCodecName(tc->encodeChoice, AVIF_CODEC_FLAG_CAN_ENCODE) || !avifCodecName(tc->decodeChoice, AVIF_CODEC_FLAG_CAN_DECODE)) {
-            ++skippedCount;
-            testCaseDestroy(tc);
-            continue;
-        }
-
-        if (!testCaseRun(tc, dataDir, AVIF_FALSE)) {
-            ++failedCount;
-        }
-        ++totalCount;
-
-        testCaseDestroy(tc);
-    }
-
-    printf("Complete. %d tests ran, %d skipped, %d failed.\n", totalCount, skippedCount, failedCount);
-
-    cJSON_Delete(tests);
-    return (failedCount == 0) ? 0 : 1;
-}
-
 typedef struct avifIOTestReader
 {
     avifIO io;
@@ -348,7 +187,8 @@ static int runIOTests(const char * dataDir)
         size_t filenameLen = strlen(filename);
         if ((ioDirLen + filenameLen) > FILENAME_MAX_LENGTH) {
             printf("Path too long: %s\n", filename);
-            return 1;
+            retCode = 1;
+            break;
         }
         strcpy(fullFilename, ioDir);
         strcat(fullFilename, filename);
@@ -356,16 +196,23 @@ static int runIOTests(const char * dataDir)
         FILE * f = fopen(fullFilename, "rb");
         if (!f) {
             printf("Can't open for read: %s\n", filename);
-            return 1;
+            retCode = 1;
+            break;
         }
         fseek(f, 0, SEEK_END);
         size_t fileSize = ftell(f);
         fseek(f, 0, SEEK_SET);
-        avifRWDataRealloc(&fileBuffer, fileSize);
+        if (avifRWDataRealloc(&fileBuffer, fileSize) != AVIF_RESULT_OK) {
+            printf("Out of memory when allocating buffer to read file: %s\n", filename);
+            fclose(f);
+            retCode = 1;
+            break;
+        }
         if (fread(fileBuffer.data, 1, fileSize, f) != fileSize) {
             printf("Can't read entire file: %s\n", filename);
             fclose(f);
-            return 1;
+            retCode = 1;
+            break;
         }
         fclose(f);
 
@@ -445,19 +292,12 @@ static int runIOTests(const char * dataDir)
 
 static void syntax(void)
 {
-    fprintf(stderr,
-            "Syntax: aviftest [options] dataDir [testFilter]\n"
-            "Options:\n"
-            "    -g : Generate Encode/Decode tests\n"
-            "    --io-only : Run IO tests only\n");
+    fprintf(stderr, "Syntax: aviftest dataDir\n");
 }
 
 int main(int argc, char * argv[])
 {
     const char * dataDir = NULL;
-    const char * testFilter = NULL;
-    avifBool generate = AVIF_FALSE;
-    avifBool ioOnly = AVIF_FALSE;
 
 #ifdef WIN32_MEMORY_LEAK_DETECTION
     _CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);
@@ -467,14 +307,10 @@ int main(int argc, char * argv[])
     // Parse cmdline
     for (int i = 1; i < argc; ++i) {
         char * arg = argv[i];
-        if (!strcmp(arg, "-g")) {
-            generate = AVIF_TRUE;
-        } else if (!strcmp(arg, "--io-only")) {
-            ioOnly = AVIF_TRUE;
+        if (!strcmp(arg, "--io-only")) {
+            fprintf(stderr, "WARNING: --io-only is deprecated; ignoring.\n");
         } else if (dataDir == NULL) {
             dataDir = arg;
-        } else if (testFilter == NULL) {
-            testFilter = arg;
         } else {
             fprintf(stderr, "Too many positional arguments: %s\n", arg);
             syntax();
@@ -496,16 +332,7 @@ int main(int argc, char * argv[])
     printf("Codec Versions: %s\n", codecVersions);
     printf("Test Data Dir : %s\n", dataDir);
 
-    int retCode = 1;
-    if (generate) {
-        retCode = generateEncodeDecodeTests(dataDir);
-    } else {
-        retCode = runIOTests(dataDir);
-        if ((retCode == 0) && !ioOnly) {
-            retCode = runEncodeDecodeTests(dataDir, testFilter);
-        }
-    }
-
+    int retCode = runIOTests(dataDir);
     if (retCode == 0) {
         printf("AVIF Test Suite: Complete.\n");
     } else {
